"""
=========
highiq.io
=========

.. currentmodule:: highiq.io

This module contains the I/O methods for loading data into and saving data from HighIQ analyses.

.. autosummary::
    :toctree: generated/

    load_arm_netcdf

"""
from .arm_data import load_arm_netcdf